from .model import WhaleVADClassifier
from .spectrogram import SpectrogramExtractor
from .weights import whalevad, WhaleVAD_Weights
from .__version__ import __version__

__all__ = [
    "WhaleVADClassifier",
    "SpectrogramExtractor",
    "whalevad",
    "WhaleVAD_Weights",
    "__version__",
]



from typing import Callable, Literal, Optional
import warnings

import torch
from torch import Tensor
from torch.nn import Module
from torchaudio.transforms import (
    Spectrogram,
)

from whalevad.utils import padding_mask, unpad_mean


class SpectrogramExtractor(Module):
    """
    Extracts spectrograms from audio signals.

    Args:
        sample_rate (int): The sample rate of the audio signals.
        n_fft (int): The number of FFT points.
        win_length (Optional[int]): The window length.
        hop_length (Optional[int]): The hop length.
        pad (int): The padding size.
        power (float | None): The power to raise the spectrogram to.
        normalized (bool | Literal["window", "frame_length"]): Whether to normalize the spectrogram.
        window_fn (Callable[..., Tensor] | str): The window function.
        center (bool): Whether to center the spectrogram.
        onesided (bool): Whether to use one-sided spectrogram.
        pad_mode (str): The padding mode.
        norm_features (Literal["demean"] | None): Whether to normalize the features.
        complex_repr (Literal["real+imag", "mag+phase", "real+imag+mag+phase", "trig"] | None): The complex representation.
        apply_log (bool): Whether to apply logarithm to the spectrogram.
    """

    def __init__(
        self,
        *,
        sample_rate: int,
        n_fft=256,
        win_length: Optional[int] = None,  # Default: n_fft
        hop_length: Optional[int] = None,  # Default: win_length//2
        pad: int = 0,
        power: float | None = 2.0,
        normalized: bool | Literal["window", "frame_length"] = False,
        window_fn: Callable[..., Tensor] | str = "hann_window",
        center: bool = False,
        onesided: bool = True,
        pad_mode: str = "reflect",
        norm_features: Literal["demean"] | None = None,
        complex_repr: (
            Literal["real+imag", "mag+phase", "real+imag+mag+phase", "trig"] | None
        ) = "real+imag",
        apply_log: bool = False,
    ) -> None:
        super().__init__()
        if win_length is not None and n_fft != win_length:
            warnings.warn(
                f"n_fft != window_len, note output rate is based on n_fft not window_len. (See torch.sftf, {n_fft} != {win_length})"
            )

        # TODO: move to fn
        n_freq_bins = n_fft // 2 + 1

        if isinstance(window_fn, str):
            window_fn = getattr(torch, window_fn)
        assert isinstance(window_fn, Callable)

        self.spectrogram = Spectrogram(
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
            pad=pad,
            power=power,
            normalized=normalized,
            window_fn=window_fn,
            center=center,
            onesided=onesided,
            pad_mode=pad_mode,
        )
        self.num_features = n_freq_bins
        self.sample_rate = sample_rate
        self.norm_features = norm_features
        self.complex_repr = complex_repr
        self.power = power
        self.apply_log = apply_log
        self.eps = 1e-8

    def forward(
        self,
        audio: Tensor,  # shape: ([batch], [channels], time)
        *,
        lengths: Optional[Tensor] = None,
        **_,
    ):  # shape: ([batch], [channels=1], time, feat)
        """
        Extracts spectrograms from audio signals.

        Args:
            audio (Tensor): The audio signals.
            lengths (Optional[Tensor]): The lengths of the audio signals.
            **_: Ignored additional arguments.

        Returns:
            Tensor: The spectrogram.
        """
        # add empty channel dim
        if audio.ndim == 1:
            audio = audio.unsqueeze(0)
        # add empty batch dim
        if audio.ndim == 2:
            audio = audio.unsqueeze(0)

        feat: Tensor = self.spectrogram(audio)
        # shape: batch, *, channel_dim=1, feat_dim, time_dim
        # Remove empty channel dim
        feat = feat.squeeze(-3)
        # ensure features are in correct order
        feat = feat.transpose(-1, -2)
        # shape: batch, time_dim, feat_dim

        if lengths is not None and self.norm_features == "demean":
            feat_mean = unpad_mean(feat, dim=-2, keepdim=True, lengths=lengths)
            # shape: (batch, 1, feat_dim)
            feat = feat - feat_mean
            pad = padding_mask(lengths)
            feat[pad] = 0.0

        elif self.norm_features == "demean":
            feat_mean = torch.mean(feat, dim=-2, keepdim=True)
            # shape: (batch, 1, feat_dim)
            feat = feat - feat_mean

        elif self.norm_features is None:
            pass
        else:
            raise ValueError(f"Unknown norm_features={self.norm_features}")

        # power=1: energy signal
        # power=2: power signal
        power_scale = 10 if self.power == 2 else 20
        if not torch.is_complex(feat) or self.complex_repr is None:
            assert self.power is not None
            # Real valued signal, power or energy
            if self.apply_log:
                feat = torch.log10(feat) * power_scale

        elif self.complex_repr == "real+imag":
            assert not self.apply_log
            feat = torch.stack([feat.real, feat.imag], dim=1)

        elif self.complex_repr == "mag+phase":
            mag = feat.abs()
            if self.apply_log:
                mag = torch.log10(mag + self.eps) * power_scale

            feat = torch.stack([mag, feat.angle()], dim=1)

        elif self.complex_repr == "real+imag+mag+phase":
            mag = feat.abs()
            if self.apply_log:
                mag = torch.log10(mag + self.eps) * power_scale

            feat = torch.stack([feat.real, feat.imag, mag, feat.angle()], dim=1)

        elif self.complex_repr == "trig":
            mag = feat.abs()
            angle = feat.angle()

            if self.apply_log:
                mag = torch.log10(mag + self.eps) * power_scale

            feat = torch.stack([mag, torch.cos(angle), torch.sin(angle)], dim=1)
        else:
            raise ValueError(f"unknown complex_repr={self.complex_repr}")
        return feat, None



from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple

import io
import torchaudio

import torch
from torch import Tensor
from torch.nn.utils.rnn import unpad_sequence


def unpad_mean(
    input: Tensor, *, dim: Optional[int] = None, keepdim: bool = False, lengths
):
    """
    Computes the mean of a sequence of tensors along a specified dimension.

    Args:
        input (Tensor): The input tensor.
        dim (Optional[int]): The dimension along which to compute the mean.
        keepdim (bool): Whether to keep the dimension of the input tensor.
        lengths (List[int] | Tensor): The lengths of the sequences.

    Returns:
        Tensor: The mean of the sequence of tensors.
    """

    input_seq = unpad_sequence(input, lengths=lengths, batch_first=True)
    if keepdim:
        input_seq = [seq.unsqueeze(0) for seq in input_seq]

    seq_mean = [torch.mean(seq, dim=dim, keepdim=keepdim) for seq in input_seq]
    return torch.vstack(seq_mean)


def padding_mask(
    lengths: List[int] | Tensor,
    *,
    dtype: torch.dtype = torch.bool,
    device: torch.device | None = None,
    max_len: Optional[int] = None,
) -> Tensor:
    """
    Creates a padding mask for a batch of sequences.

    Args:
        lengths (List[int] | Tensor): The lengths of the sequences.
        dtype (torch.dtype): The data type of the mask.
        device (torch.device | None): The device on which to create the mask.
        max_len (Optional[int]): The maximum length of the sequences.

    Returns:
        Tensor: The padding mask.
    """
    if isinstance(lengths, Tensor):
        lengths = lengths.tolist()
    batch_size = len(lengths)
    time = max_len if max_len is not None else max(lengths)

    mask = torch.zeros((batch_size, time), dtype=dtype, device=device)

    for i, l in enumerate(lengths):
        assert l <= time, "index of bounds, increate max_len"
        mask[i, l:] = True
    return mask


def load_remote_audio(
    url: str,
    filename: str,
    *,
    frame_offset: Optional[int] = None,
    num_frames: Optional[int] = None,
    start_s: Optional[float] = None,
    end_s: Optional[float] = None,
    sample_rate: Optional[int] = None,
    **kwargs,
):
    from remotezip import RemoteZip

    if start_s is not None:
        assert frame_offset is None, "offset and start_s cannot be both specified"
        assert (
            sample_rate is not None
        ), "sample_rate must be specified when using start_s"
        frame_offset = int(start_s * sample_rate)
    frame_offset = frame_offset or 0
    del start_s

    if end_s is not None:
        assert num_frames is None, "num_frames and end_s cannot be both specified"
        assert sample_rate is not None, "sample_rate must be specified when using end_s"
        end = int(end_s * sample_rate)
        num_frames = end - frame_offset
        assert num_frames > 0, "num_frames must be positive"
    num_frames = num_frames or -1
    del end_s

    with RemoteZip(url) as zf:
        with zf.open(filename) as file:
            buffer = io.BytesIO(file.read())
            audio, sr = torchaudio.load(
                buffer,
                frame_offset=frame_offset,
                num_frames=num_frames,
                **kwargs,
            )
    assert (
        sr == sample_rate
    ), f"sample rate mismatch: {sr}Hz (file) != {sample_rate}Hz (expected)"
    return audio, sr


@dataclass
class ATBFLExemplar:
    filename: str
    start_s: float
    end_s: float
    sample_rate: int = 250

    def fetch_and_load(self, **kwargs):
        return get_atbfl_exemplar(
            self.filename,
            start_s=self.start_s,
            end_s=self.end_s,
            sample_rate=self.sample_rate,
            **kwargs,
        )


ATBFL_REPO_URLS = {
    "train": "https://zenodo.org/records/15092732/files/biodcase_development_set.zip?download=1",
    "val": "https://zenodo.org/records/15092732/files/biodcase_development_set.zip?download=1",
}
ATBFL_EXAMPLARS = {
    "train": ATBFLExemplar(
        "biodcase_development_set/train/audio/kerguelen2005/2005-06-21T18-00-00_000.wav",
        960.0,
        970.0,
    ),
    "val": ATBFLExemplar(
        "biodcase_development_set/validation/audio/kerguelen2014/2014-06-29T23-00-00_000.wav",
        2755.0,
        2765.0,
    ),
}


def get_atbfl_exemplar(
    filename: Optional[str] = None,
    *,
    split: Literal["train", "val"] | None = None,
    **kwargs,
) -> Tuple[Tensor, int]:
    """
    Fetches and loads an ATBFL exemplar from the repository. This operation only downloads the required chunks from the repo not the entire zip file.

    Args:
        filename: The filename of the audio file to load.
        start: The start time in seconds.
        end: The end time in seconds.
        split: The split to use ("train" or "val")

    Returns:
        A tuple containing the audio tensor and the sample rate.
    """
    if filename is None:
        split = split or "train"
        exemplar = ATBFL_EXAMPLARS[split]

        return exemplar.fetch_and_load(**kwargs, split=split)
    assert split is not None, "split must be specified"

    return load_remote_audio(ATBFL_REPO_URLS[split], filename, **kwargs)



from typing import Callable, Dict, List, Literal, Optional, Tuple

from torch import Tensor, adaptive_avg_pool1d
import torch
from torch.nn import (
    GELU,
    LSTM,
    BatchNorm2d,
    Conv2d,
    Flatten,
    Linear,
    MaxPool2d,
    Module,
    ModuleList,
    Sequential,
    Dropout,
    Dropout2d,
)
from torch.nn.utils.rnn import (
    pad_packed_sequence,
    pack_padded_sequence,
)


class WhaleVADClassifier(Module):
    """
    Create Whale-VAD classifier.

    Args:
        num_classes (Optional[int]): The number of classes to classify.
        feat_channels (int): The number of input feature channels.
        include_intermediate_features (bool): Whether to include intermediate features.
        include_bounding_boxes (bool): Whether to include bounding boxes.
        include_bottleneck_layers (bool): Whether to include bottleneck layers.
        include_aggregation_layers (bool): Whether to include aggregation layers.
        num_anchors (int): The number of anchors.
        return_hidden_state (bool): Whether to return the hidden state.
    """

    _class_mapping = ["bmabz", "d", "bp"]

    def __init__(
        self,
        *,
        num_classes: Optional[int] = None,
        feat_channels=1,
        include_intermediate_features: bool = False,
        include_bounding_boxes: bool = False,
        include_bottleneck_layers: bool = True,
        include_aggregation_layers: bool = True,
        num_anchors: int = 64,
        return_hidden_state: bool = False,
    ) -> None:
        super().__init__()
        num_classes = (
            num_classes if num_classes is not None else len(self._class_mapping)
        )
        self.num_anchors = num_anchors
        self.return_hidden_state = return_hidden_state

        # Act as a learnable mel
        self.fbank = Conv2d(
            in_channels=feat_channels,
            out_channels=64,
            kernel_size=(7, 1),  # only convolve over spectrum not time
            stride=(3, 1),
            padding=0,  # we can drop some spec features
        )

        conv_block_feat_extractor = Sequential(
            # LAYER 1
            Conv2d(
                in_channels=64,
                out_channels=128,
                kernel_size=(5, 5),
                stride=(3, 1),
                padding=5 // 2,  # needed to keep output resolution the same
            ),
            BatchNorm2d(128),
            GELU(),
            MaxPool2d(
                kernel_size=(5, 1),
                stride=1,
                padding=0,
            ),
            # ------------------
            # LAYER 2
            Conv2d(
                in_channels=128,
                out_channels=128,
                kernel_size=(3, 3),
                stride=(2, 1),
                padding=3 // 2,
            ),
            BatchNorm2d(128),
            GELU(),
            MaxPool2d(
                kernel_size=(3, 1),
                stride=1,
                padding=0,
            ),
        )
        conv_block_bottleneck = Sequential(
            # ------------------
            # LAYER 3
            Conv2d(
                in_channels=128,
                out_channels=64,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=1 // 2,
            ),
            GELU(),
            Dropout(0.1),
            # ------------------
            # LAYER 4
            Conv2d(
                in_channels=64,
                out_channels=64,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=3 // 2,
            ),
            GELU(),
            Dropout(0.1),
            # ------------------
            # LAYER 5
            Conv2d(
                in_channels=64,
                out_channels=128,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=1 // 2,
            ),
            BatchNorm2d(128),
            GELU(),
            Dropout(0.1),
        )

        conv_block_feat_agg = Sequential(
            # Spatial dropout: https://arxiv.org/pdf/1411.4280
            Dropout2d(0.2),
            # ------------------
            # LAYER 6
            Conv2d(
                in_channels=128,
                out_channels=128,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=3 // 2,
                groups=128,
            ),
            BatchNorm2d(128),
            GELU(),
            # ------------------
            # LAYER 7
            Conv2d(
                in_channels=128,
                out_channels=128,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=3 // 2,
                groups=128,
            ),
            BatchNorm2d(128),
            GELU(),
            # ------------------
            # LAYER 8
            Conv2d(
                in_channels=128,
                out_channels=128,
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=3 // 2,
                groups=128,
            ),
            BatchNorm2d(128),
            GELU(),
        )

        cnn_blocks: List[Module] = []
        if include_bottleneck_layers:
            cnn_blocks.append(conv_block_bottleneck)

        if include_aggregation_layers:
            cnn_blocks.append(conv_block_feat_agg)
        # Residual blocks
        # TODO: make residule module
        self.cnn_blocks = Sequential(
            conv_block_feat_extractor,
            ResidualBlock(  # Expected residual shape: (batch, channels=128, feat=3, time)
                *cnn_blocks,
                output_residuals=include_intermediate_features,
            ),
        )
        # flatten to shape (batch, time, channel*feat=3*128)

        # Eigen dim reduction
        if include_intermediate_features:
            # three residual/intermediate layer with 3 channels each
            # num_res*num_chan*num_feat=3*128*3
            self.feat_proj = Linear(3 * 128 * 3, 64)
            self.bb_proj = Linear(3 * 128 * 3, 64)
        else:
            # channels*num_features=128*3
            self.feat_proj = Linear(128 * 3, 64)
            self.bb_proj = Linear(128 * 3, 64)

        self.lstm = LSTM(
            input_size=64,  # output of feat_proj
            hidden_size=128,
            bidirectional=True,
            num_layers=2,
            dropout=0.5,
        )

        self.classifier = Linear(
            in_features=128 * 2,  # hidden_size * 2 (bidirectional)
            out_features=num_classes,
        )

        self.include_bounding_boxes = include_bounding_boxes
        self.bounding_box_mlp = (
            Sequential(
                Linear(64, 128),
                Dropout(0.05),
                GELU(),
                Linear(128, 128),
                Dropout(0.05),
                GELU(),
            )
            if include_bounding_boxes
            else None
        )
        self.bounding_box_reg = (
            Linear(128, 4) if include_bounding_boxes else None  # ([XYXY])
        )

        self.bounding_box_conf = (
            Sequential(
                Linear(128, 1),
                Flatten(),
            )
            if include_bounding_boxes
            else None
        )

    def forward(
        self,
        features: Tensor,  # shape(batch_dim, *, [channel], time, embedding_size)
        lab_lengths: Optional[Tensor] = None,
        hidden_state: Optional[Tensor] = None,
        **opts,
    ) -> Tuple[Tensor, Tensor, Dict]:
        """Create Whale-VAD classifier.

        Args:
            features (Tensor): Input features.
            lab_lengths (Optional[Tensor]): Lengths of labels.
            hidden_state (Optional[Tensor]): Hidden state.
            **opts: Additional options.

        Returns:
            Tuple[Tensor, Tensor, Dict]: Output logits, frame-level call probabilities and additional internal state.
        """
        # If lab_lengths is None, assume no padding
        if lab_lengths is None:
            batch_size = features.size(0) if features.ndim >= 4 else 1
            time_dim = features.size(-2)
            lab_lengths = torch.full((batch_size,), time_dim)

        return_opts = dict()
        # Swap time and feature dim
        features = features.transpose(-1, -2)
        # add empty channel dim
        if features.ndim < 4:
            features = features.unsqueeze(1)
        # shape: (batch, chanel=1, dim, frames)

        # Think mel...
        filt_feat = self.fbank(features)
        # shape: (batch, channel=64, features=58, time)

        # Think cepstrum
        cepstrum = self.cnn_blocks(filt_feat)
        # shape: (batch, channel=128, features=3, time)
        cepstrum = cepstrum.flatten(start_dim=1, end_dim=2)
        cepstrum = cepstrum.transpose(-1, -2)
        # shape: (batch, time, channel*feat=384)

        # Down project
        cepstrum_proj = self.feat_proj(cepstrum)
        # shape: (batch, time, proj=128)

        # Create packed sequence with padding already applied
        lab_lengths = lab_lengths.cpu()
        feat_seq = pack_padded_sequence(
            cepstrum_proj,
            lab_lengths,
            batch_first=True,
            enforce_sorted=False,  # Could drop since collate ensures it is sorted
        )

        Z_seq, hidden_state = self.lstm(feat_seq, hidden_state)

        # Convert packed sequence to regular tensor with zero padding
        Z, Z_len = pad_packed_sequence(Z_seq, batch_first=True)
        assert torch.all(Z_len == lab_lengths)

        return_opts["rnn_lengths"] = Z_len
        return_opts["hidden_state"] = (
            hidden_state if self.return_hidden_state else None,
        )

        logits = self.classifier(Z)
        pred = torch.nn.functional.sigmoid(logits)

        if self.include_bounding_boxes:
            assert self.bounding_box_conf is not None
            assert self.bounding_box_reg is not None
            assert self.bounding_box_mlp is not None

            bb_latent = self.bb_proj(cepstrum)
            bb_latent = bb_latent.transpose(-1, -2)
            # shape: batch, proj, time
            bb_latent = adaptive_avg_pool1d(bb_latent, self.num_anchors)
            bb_latent = bb_latent.transpose(-1, -2)
            # shape: batch, num_anchors, proj
            bb_latent = self.bounding_box_mlp(bb_latent)

            bounding_box_conf = self.bounding_box_conf(bb_latent)
            bounding_box = self.bounding_box_reg(bb_latent)

            return_opts["bounding_box_reg"] = bounding_box
            return_opts["bounding_box_conf"] = bounding_box_conf
        # shape: (batch, time, proj=128)

        return logits, pred, return_opts


class ResidualBlock(Module):
    def __init__(
        self,
        *blocks: Module,
        connection: Literal["parallel", "series"] = "series",
        residual: Literal["concat", "sum"] = "sum",
        output_residuals: bool = False,
    ):
        super().__init__()
        assert residual == "sum"
        assert connection == "series"

        self.blocks = ModuleList(blocks)
        self.output_residuals = output_residuals

    def forward(self, X: Tensor):
        if self.output_residuals:
            residuals = [X]

        for block in self.blocks:
            X = X + block(X)

            if self.output_residuals:
                residuals.append(X)

        if self.output_residuals:
            return torch.concat(residuals, dim=-2)
        return X


class WhaleVADModel(Module):
    """
    Construct complete Whale-VAD model, containing **both** classifier and feature extractor (transform).

    Args:
        classifier: WhaleVADClassifier
        transform: Optional[Module | Callable[[Tensor], Tensor]]
    """

    def __init__(
        self,
        classifier: WhaleVADClassifier,
        transform: Optional[Module | Callable[[Tensor], Tensor]],
    ):
        super().__init__()
        self.classifier = classifier
        self.transform = transform

    def forward(self, audio: Tensor, **opts):
        spec = audio
        if self.transform:
            spec, trans_opts = self.transform(audio)
            opts.update(**opts)
        return self.classifier(spec, **opts)

    def __iter__(self):
        yield self.classifier
        yield self.transform



from enum import Enum
from typing import Any, Callable, Dict, Mapping, Optional, Self
from dataclasses import dataclass, field

from torch import Tensor
from torch.nn import Module
from torch.hub import load_state_dict_from_url

from whalevad.model import WhaleVADClassifier, WhaleVADModel
from whalevad.spectrogram import SpectrogramExtractor

__all__ = [
    "WhaleVAD_Weights",
    "whalevad",
]


# Based on https://github.com/pytorch/vision/blob/d5df0d67dc43db85a3963795903b51c57a6146c1/torchvision/models/_api.py
@dataclass
class Weights:
    """
    Dataclass containing configuration related to model weights.

    Args:
        url (str): The URL to download the pre-trained weights from.
        transform (Optional[Callable | Module]): The transformation to apply to the input data.
        meta (Dict[str, Any]): Additional metadata about the weights.
        model_config (Dict[str, Any]): Configuration parameters for the model.
    """

    url: str
    transform: Optional[Callable | Module] = None
    meta: Dict[str, Any] = field(default_factory=dict)
    model_config: Dict[str, Any] = field(default_factory=dict)


class WeightsEnum(Enum):
    """
    This class is the parent class of all model weights. Each model building method receives an optional `weights`
    parameter with its associated pre-trained weights. It inherits from `Enum` and its values should be of type
    `Weights`.

    Args:
        value (Weights): The data class entry with the weight information.
    """

    value: Weights  # type: ignore

    def get_state_dict(self, *args: Any, **kwargs: Any) -> Mapping[str, Any]:
        return load_state_dict_from_url(self.url, *args, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self._name_}"

    @property
    def url(self):
        return self.value.url

    @property
    def transform(self):
        return self.value.transform

    @property
    def meta(self):
        return self.value.meta

    @property
    def model_config(self):
        return self.value.model_config

    @classmethod
    def verify(cls, inst: str | Self) -> Self:
        if isinstance(inst, WeightsEnum):
            return inst

        return getattr(cls, inst)


class WhaleVAD_Weights(WeightsEnum):
    ATBFL_DCASE_3P_V1 = Weights(
        url="https://github.com/CMGeldenhuys/Whale-VAD/releases/download/v0.1.0/WhaleVAD_ATBFL_3P-c6f6a07a.pt",
        model_config=dict(
            num_classes=7,
            feat_channels=3,
        ),
        transform=SpectrogramExtractor(
            sample_rate=250,
            n_fft=256,
            win_length=256,
            hop_length=5,
            norm_features="demean",
            power=None,
            complex_repr="trig",
        ),
    )
    DEFAULT = ATBFL_DCASE_3P_V1


def whalevad(
    weights: Optional[WhaleVAD_Weights | str] = None,
    progress: bool = True,
    transform: Optional[Module | Callable] = None,
    eval: bool = True,
    **kwargs,
) -> WhaleVADModel:
    """
    Create a WhaleVAD model with pre-trained weights.

    Args:
        weights (Optional[WhaleVAD_Weights | str]): The weights to use for the model.
        progress (bool): Whether to show a progress bar while downloading the weights.
        transform (Optional[Module | Callable]): The transformation to apply to the input data.
        **kwargs: Additional keyword arguments to pass to the model constructor.

    Returns:
        WhaleVADModel

    Example:
        >>> model = whalevad(weights='DEFAULT')
        >>> model.eval()
        >>> classifier, transform = model
    """
    if weights is None:
        clf = WhaleVADClassifier(**kwargs)
        return WhaleVADModel(clf, transform)
    weights = WhaleVAD_Weights.verify(weights)
    state = weights.get_state_dict(progress=progress, check_hash=True)
    clf = WhaleVADClassifier(**weights.model_config, **kwargs)
    clf.load_state_dict(state)

    if transform is None:
        transform = weights.transform

    model = WhaleVADModel(clf, transform)

    if eval:
        model = model.eval()

    return model
